-- Optimal Portfolio Balancer 
--
-- Computes statistics for investment funds, stocks, etc.
-- and chooses the optimal mix for each level of desired risk/return.
-- 
-- This currently uses weekly data for Bank of Montreal mutual funds,
-- but you could adapt it to your own portfolio of stocks, bonds, 
-- mutual funds, etc.
--
-- Written by Rob Craig, RDS
-- http://www.RapidEuphoria.com
-- 
--   Disclaimer: This is free and Public Domain. Use it at your own risk!
--   If you lose money, I will not be held responsible! 
--   If you gain money, feel free to donate some of it. :-)

-- Usage (in a command-prompt window):
--
--        exwc bmocor > temp
--
-- Sorry, there is no nice user interface for this.
-- I just edit the data weekly, and run the program for a few hours,
-- crunching a few hundred million combinations. It runs nicely on
-- the second CPU of my new dual-core computer.
--
-- It will run significantly faster if you Translate/Compile to bmocor.exe,
-- using the Euphoria to C Translator, and run: bmocor.exe > temp

-- What it does:
--
--   1. Computes averages and standard deviations for each fund.
--      This gives you an idea of how risky each fund is.
--
--   2. Computes correlations between all funds (-1 to +1).
--      If fund A is positively correlated with fund B, it
--      means that the two tend to go up or down together.
--      If they are negatively correlated, they tend to move
--      in opposite directions. If the correlation is zero, 
--      they move purely at random compared with each other.

--      For maximum diversification and safety, while maintaining a
--      high expected rate of return, you want to match up funds 
--      that are negatively correlated, or at least have only a small 
--      positive correlation. 
--
--      Every investor would like to make a steady 8% return each year, 
--      rather than gamble their life savings 50/50
--      on say an 8+20 (28%) or an 8-20 (-12%) return. With 28%
--      you would waste the money on silly luxuries. With -12% you
--      might have to cut your food budget, or go out and get a job! :-)
--
--   3. Goes through every possible combination of funds, keeping the
--      combos that "make sense", i.e. they offer a higher return
--      or lower risk than all other combos. A list is shown of
--      all good combinations, along with your current combination.
--      You might want to adjust your current combination to one that
--      offers greater return for the same or less risk, or less 
--      risk for the same or greater return.


-- Section 1: Data you need to edit each period (week):

-- Raw files containing the fund values
sequence files
files = {
    "20070608",
    "20070615",
    "20070622",
    "20070629",
    "20070706",
    "20070713",
    "20070720",
    "20070727",
    "20070803",
    "20070810",
    "20070817",
    "20070824",
    "20070831",
    "20070907",
    "20070914",
    "20070921",
    "20070928",
    "20071005",
    "20071012",
    "20071019",
    "20071026",
    "20071102",
    "20071109"
    
    --,"latest"
}

-- Value of a guaranteed rate of return fund:
constant GIC = 20000 -- fixed income (GIC's, bank account interest etc.)

-- Unit values (number of shares etc.) that you currently hold in your funds
sequence your_funds -- lower case
your_funds = {  -- I just made up all these fund unit numbers!
		-- Not my actual portfolio!
    {"equity fund", 1000},  
    {"special equity fund", 800},
    {"japanese", 2500},
    {"greater china class", 1500},
    {"global high yield", 5000},
    {"precious metals", 100},
    {"gic", GIC}
}

-- Funds to consider for the optimization algorithm -- lower case
sequence possible_funds 
constant NAME = 1,
	 EXPECTED_RETURN = 2, -- yearly expected return
	 MIN_TRY = 3, -- minimum percentage to consider
	 MAX_TRY = 4, -- maximum percentage to consider
	 YOUR_VAL = 5, -- added later
	 SERIES = 6 -- aded later
-- The MIN and MAX percentages could initially be 0 to 100 for all funds,
-- but to save time, as you decrease the STEP size,
-- you will probably want to decrease the range of values on some funds.
	 
-- Step size - be careful, a small value will cause a huge increase
-- in the number of combinations to test. 
-- Best if it's an nice integer, 1,2,4,5,10,...
constant STEP = 10  -- percent (so you won't have to wait hours for this to finish!)

-- Funds that you want the alogorithm to optimize.
-- You must fill in the funds name, expected (average) rate of return,
-- and the minimum and maximum percentage to consider for this fund
-- in your portfolio.

-- Where do you get the values to plug in for "expected rate of return" ????
-- You just have to estimate it, based on past history,
-- and your "gut feel" for what is likely to happen in the future.
-- Generally, the more volatile a fund is, the higher is the expected
-- rate of return.

possible_funds = {
--    NAME                   RR  MIN MAX   
   {"gic",                   4.5, 0, 100},
   {"global science & technology", 4.0, 0, 100},
   {"japanese",              5.9, 0,  100}, 
     -- we expect Japanese fund to earn just 5.9%
     -- and we will test all combos from 0 to 100% Japanese 
   {"international index",   6.5, 0, 100},
   {"international equity",  6.5, 0, 100},
   {"european",              6.3, 0, 100},
   {"emerging markets",      7.8, 0, 100},
   {"special equity fund",   7.4, 0, 100},
   {"u.s. equity index",     5.5, 0, 100},
   {"u.s. growth",           5.8, 0, 100},
   {"greater china class",   7.3, 0, 100}, 
   {"resource fund",         7.7, 0, 100},
   {"precious metals fund",  8.3, 0, 100},
   {"world bond fund",       3.0, 0, 100},
   {"global high yield",     4.0, 0, 100}, 
   {"bond fund",             3.7, 0, 100}, 
   {"equity fund",           6.8, 0, 100}
}


--------------------------------------------------------------------------
-- Section 2 - You will rarely have to edit the code below 
-- (after setting it up initially to parse your raw input data)

include get.e
include misc.e
include wildcard.e
include sort.e


-- Basic Statistical functions

function sum(sequence s)
-- sum 
    atom t
     
    t = 0
    for i = 1 to length(s) do
	t += s[i]
    end for 
    return t
end function

function avg(sequence x)
-- average
    return sum(x)/length(x)
end function

function std_dev(sequence data)
   -- Fast version. Thanks to Juergen Luethje
   atom x, sum_q, sum_x
   
   sum_q = 0
   sum_x = 0
   for i = 1 to length(data) do
      x = data[i]
      sum_q += x*x
      sum_x += x
   end for
   return sqrt( (sum_q - sum_x*sum_x/length(data)) / length(data) )
end function

function cor(sequence x, sequence y)
-- linear correlation coefficient, -1 to +1
    atom avgx, avgy, devx, devy, cov
    
    if length(y) > length(x) then
	y = y[$-length(x)+1..$]
    elsif length(x) > length(y) then
	x = x[$-length(y)+1..$]
    end if
    
    avgx = sum(x)/length(x)
    avgy = sum(y)/length(y)
    devx = std_dev(x)
    devy = std_dev(y)
    cov  = sum((x - avgx) * (y - avgy)) / length(x)
    if devx*devy = 0 then
	return 0
    else
	return cov/(devx*devy)
    end if
end function

constant TRUE = 1, FALSE = 0

type boolean(integer x)
    return x = TRUE or x = FALSE
end type

-- first column:
for i = 1 to length(possible_funds) do
    possible_funds[i] =  append(possible_funds[i], 0) -- for your mix
end for

sequence all_fund_prices, all_fund_names
all_fund_prices = {}
all_fund_names = {}

function is_your_fund(sequence name)
-- Is name a fund you have money in?
    for i = 1 to length(your_funds) do
	if match(lower(your_funds[i][1]), lower(name)) = 1 then
	    return i
	end if
    end for
    return 0
end function


-- Read in all the raw data files and create the time series of fund values

object line
sequence v, name
integer f, fund, p
boolean started, found_your

all_fund_names = {}
for i = 1 to length(files) do
    -- read next weekly file
    f = open("bmo_" & files[i] & ".txt", "r")
    started = FALSE
    fund = 1
    
    while 1 do
	-- read next line of weekly file
	line = gets(f)
	if atom(line) then
	    puts(1, "got to end of file with no u.s. dollar!\n")
	    exit
	end if
    
	if match("Security Funds", line) then
	    started = TRUE
	end if
	
	if started and match("BMO ", line) then
	    -- record name and price
	    line = line[5..$]
	    p = find('\t', line)
	    if p = 0 then
		puts(1, "missing tab!\n")
		exit
	    end if
	    name = line[1..p-1]
	    if length(name) and name[$] = ' ' then
		name = name[1..$-1]
	    end if
	    if i = 1 then
		-- first file only
		all_fund_names = append(all_fund_names, name)
		all_fund_prices = append(all_fund_prices, {})
	    end if
	    line = line[p+2..$]
	    v = value(line)
	    if v[1] = GET_SUCCESS then
		all_fund_prices[fund] = append(all_fund_prices[fund], v[2])
		fund += 1
	    else
		puts(1, "couldn't read" & files[i] & '\n')
		abort(1)
	    end if
	    
	elsif match("U.S. Dollar Funds", line) then
	    exit
	end if
    
    end while
    
    close(f)
end for
puts(1, '\n')

-- Add GIC as a simulated fund, value 1, with zero fluctuations
all_fund_names = append(all_fund_names, "gic")
all_fund_prices = append(all_fund_prices, repeat(1, length(all_fund_prices[1])))

-- Display all the raw data

printf(1, "Raw Data (%d weeks)\n\n", length(all_fund_prices[1]))
for i = 1 to length(all_fund_prices) do
    printf(1, "%35s: ", {all_fund_names[i]}) 

    for j = 1 to length(all_fund_prices[i]) do
	printf(1, "%8.4f ", all_fund_prices[i][j])
    end for

    puts(1, '\n')
end for

puts(1, '\n')

-- Show the % gain or loss per time period (week)

printf(1, "Percent Gain Per Week with Standard Deviation - %s to %s\n\n", {files[1], files[$]})

atom your_total, delta, fval

your_total = 0
for i = 1 to length(all_fund_prices) do
    delta = all_fund_prices[i][$]/all_fund_prices[i][1] - 1
    printf(1, "%35s: %6.2f%% +/-%5.2f%%", 
       {all_fund_names[i], delta*100 / length(all_fund_prices[i]), 
	std_dev(100 * all_fund_prices[i][2..$] / 
		      all_fund_prices[i][1..$-1])})
    
    found_your = FALSE
    for j = 1 to length(your_funds) do
	if match(lower(your_funds[j][1]), lower(all_fund_names[i])) = 1 then
	    printf(1, "%8.0f", 
	    (all_fund_prices[i][$] * your_funds[j][2]))
	    fval = all_fund_prices[i][$] * your_funds[j][2]
	    your_total += fval
	    found_your = TRUE
	    for k = 1 to length(possible_funds) do
		if match(possible_funds[k][NAME], lower(all_fund_names[i])) = 1 then
		    possible_funds[k][YOUR_VAL] = fval
		end if
	    end for
	end if
    end for
    if not found_your then
	puts(1, repeat(' ', 8))
    end if
    
    puts(1, '\n')
end for

printf(1, "\n      TOTAL CURRENT VALUE OF YOUR FUNDS: %9.0f (+ GIC = %6.0f)\n", 
       {your_total - GIC, your_total})
puts(1, '\n')


-- Compute the cross correlations between all funds

atom c
atom weighted_cor, total_other, value
sequence sorted_cors
integer rf

puts(1, "Correlations\n")
for i = 1 to length(all_fund_prices) do
    sorted_cors = {}
    weighted_cor = 0
    total_other = 0
    for j = 1 to length(all_fund_prices) do
	c = cor(all_fund_prices[i], all_fund_prices[j])
	if i != j and c != 0 then
	    sorted_cors = append(sorted_cors, {c, all_fund_names[j]})
	end if
	rf = is_your_fund(all_fund_names[j]) 
	if rf then 
	    value = your_funds[rf][2] * all_fund_prices[j][$]
	    weighted_cor += c * value
	    total_other +=  value
	end if
    end for
    
    sorted_cors = sort(sorted_cors)
    
    for j = 1 to length(sorted_cors) do
	printf(1, "%s vs %s: %.2f\n", 
		{all_fund_names[i], sorted_cors[j][2], sorted_cors[j][1]})
	
    end for
    if weighted_cor != 0 then
	printf(1, "weighted overall correlation vs your current funds: %d\n", 100 * weighted_cor/total_other)
    end if
    puts(1, '\n')
end for

-- Compute your current mix

sequence yourmix
atom yourval

yourmix = {}
for i = 1 to length(possible_funds) do
    yourval = possible_funds[i][YOUR_VAL] * 100 / your_total
    yourmix = append(yourmix, yourval)
end for


-- Normalize the input data price series for each possible fund

for i = 1 to length(possible_funds) do
    for j = 1 to length(all_fund_prices) do
	if match(possible_funds[i][NAME], lower(all_fund_names[j])) = 1 then
	    possible_funds[i] = append(possible_funds[i], 
				all_fund_prices[j] / all_fund_prices[j][1])
	    exit
	end if
    end for
end for

-- Find the optimal mix for each level of return.
-- i.e. if a combination has less return and higher risk (standard deviation)
-- than another one, throw it out.

atom combo_count
combo_count = 0

integer dry_run

function mixture(sequence weights)
-- create a blended series, given the percentage weights    
    sequence mix

    mix = repeat(0, length(possible_funds[1][SERIES]))
    for i = 1 to length(possible_funds) do
	mix += weights[i] * possible_funds[i][SERIES]
    end for
    return mix
end function

-- the best combos
sequence best
best = {}

procedure record(sequence ret_sd, sequence mix)
-- record this combo, if it's a good one
    integer i
    atom r, sd
	 
    r = ret_sd[1]
    sd = ret_sd[2]
    i = 1
    while i <= length(best) do
	if best[i][1][1] >= r then
	   if best[i][1][2] < sd then
	       return
	   end if      
	else
	   if best[i][1][2] > sd then
	       best = best[1..i-1] & best[i+1..$] -- delete it
	       i -= 1
	   end if
	end if
	i += 1
    end while 
    best = prepend(best, {ret_sd, mix})
end procedure

procedure all_combos(sequence weights, integer f, atom total, atom used_so_far)
-- recursively go through every possible combination    
    atom sd, ret
    sequence m
    integer min_limit, max_limit
    
    max_limit = 100 - used_so_far
    if f = length(possible_funds) then
	if max_limit <= possible_funds[f][MAX_TRY] and 
	   max_limit >= possible_funds[f][MIN_TRY] then
	    combo_count += 1
	    if not dry_run then
		weights[f] = max_limit
		m = mixture(weights)
		sd = std_dev((m[2..$]/m[1..$-1]))
		ret = 0
		for j = 1 to length(possible_funds) do
		    ret += possible_funds[j][EXPECTED_RETURN] * weights[j]
		end for
		record({ret, sd}, weights)
	    end if   
	end if
    else
	min_limit = possible_funds[f][MIN_TRY]
	if max_limit > possible_funds[f][MAX_TRY] then
	    max_limit = possible_funds[f][MAX_TRY]
	end if
	for i = min_limit to max_limit by STEP do
	    if remainder(combo_count, 100000) = 0 then
		printf(2, "%.1f million\r", combo_count/1000000)
		if get_key() = 'q' then
		    abort(0)
		end if
	    end if
	    weights[f] = i
	    all_combos(weights, f+1, total-i, used_so_far + i)
	end for
    end if
end procedure

sequence m
atom yoursd, yourret

m = mixture(yourmix)
yoursd = std_dev((m[2..$]/m[1..$-1]))
yourret = 0
for j = 1 to length(possible_funds) do
    yourret += possible_funds[j][EXPECTED_RETURN] * yourmix[j] 
end for

-- Figure out how many combos we have to evaluate
puts(2, "Pre-calculate the number of combos to be tested:\n")
puts(2, "(If this number gets too high, hit q to quit)\n")

dry_run = TRUE
all_combos(repeat(0, length(possible_funds)), 1, 100, 0)

-- Now actually go through all the combos
combo_count = 0
puts(2, "\n\nNow Testing ... (hit q to abort)\n")
dry_run = FALSE
all_combos(repeat(0, length(possible_funds)), 1, 100, 0)

best = append(best, {{yourret, yoursd}, yourmix})
best = sort(best)

-- Display the results
puts(1, "Look for \"Your Current Portfolio\" below...\n\n")
for i = 1 to length(best) do
    printf(1, "return = %.3f, risk (std_dev) = %.3f", best[i][1] / {100, .01})
    if equal(best[i][2], yourmix) then
	puts(1, " <----- Your Current Portfolio")
    end if
    puts(1, '\n')
    for j = 1 to length(best[i][2]) do
	if best[i][2][j] != 0 then
	    printf(1, "%.20s %.0f\n", {possible_funds[j][NAME], best[i][2][j]})
	end if
    end for
    puts(1, '\n')
end for


